"""
Module: types

This module provides a suite of type validation functions to ensure we conform to expected types.
These functions validate whether a given object qualifies as an:
extractor, multi-extractor, loader, step, or context.
Using these type-checking functions, Pipeline can enforce proper usage of its components.
"""
from .type_validation import is_extractor
from .type_validation import is_multiextractor
from .type_validation import is_loader
from .type_validation import is_step
from .type_validation import is_context
from .type_validation import is_context_object
